@@ -21,6 +21,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <list>
+#include <cstring>
 
 #include "cryptoUtils.h"
 #include "bctoolbox/crypto.hh"
@@ -74,6 +75,11 @@ uint8_t bzrtpUtils_getAllAvailableCryptoTypes(uint8_t algoType, uint8_t availabl
 		availableTypes[index] = ZRTP_KEYAGREEMENT_DH3k;
 		index++;
 
+		if (available_key_agreements&BCTBX_KEM_MLKEM512) {
+			availableTypes[index] = ZRTP_KEYAGREEMENT_MLK1;
+			index++;
+		}
+
 		if (available_key_agreements&BCTBX_KEM_KYBER512) {
 			availableTypes[index] = ZRTP_KEYAGREEMENT_KYB1;
 			index++;
@@ -84,6 +90,11 @@ uint8_t bzrtpUtils_getAllAvailableCryptoTypes(uint8_t algoType, uint8_t availabl
 			index++;
 		}
 
+		if (available_key_agreements&BCTBX_KEM_MLKEM768) {
+			availableTypes[index] = ZRTP_KEYAGREEMENT_MLK2;
+			index++;
+		}
+
 		if (available_key_agreements&BCTBX_KEM_KYBER768) {
 			availableTypes[index] = ZRTP_KEYAGREEMENT_KYB2;
 			index++;
@@ -94,6 +105,11 @@ uint8_t bzrtpUtils_getAllAvailableCryptoTypes(uint8_t algoType, uint8_t availabl
 			index++;
 		}
 
+		if (available_key_agreements&BCTBX_KEM_MLKEM1024) {
+			availableTypes[index] = ZRTP_KEYAGREEMENT_MLK3;
+			index++;
+		}
+
 		if (available_key_agreements&BCTBX_KEM_KYBER1024) {
 			availableTypes[index] = ZRTP_KEYAGREEMENT_KYB3;
 			index++;
@@ -119,6 +135,11 @@ uint8_t bzrtpUtils_getAllAvailableCryptoTypes(uint8_t algoType, uint8_t availabl
 			index++;
 		}
 
+		if ((available_key_agreements&BCTBX_KEM_X25519) && (available_key_agreements&BCTBX_KEM_MLKEM512)) {
+			availableTypes[index] = ZRTP_KEYAGREEMENT_K255_MLK512;
+			index++;
+		}
+
 		if ((available_key_agreements&BCTBX_KEM_X25519) && (available_key_agreements&BCTBX_KEM_KYBER512)) {
 			availableTypes[index] = ZRTP_KEYAGREEMENT_K255_KYB512;
 			index++;
@@ -129,6 +150,11 @@ uint8_t bzrtpUtils_getAllAvailableCryptoTypes(uint8_t algoType, uint8_t availabl
 			index++;
 		}
 
+		if ((available_key_agreements&BCTBX_KEM_X448) && (available_key_agreements&BCTBX_KEM_MLKEM1024)) {
+			availableTypes[index] = ZRTP_KEYAGREEMENT_K448_MLK1024;
+			index++;
+		}
+
 		if ((available_key_agreements&BCTBX_KEM_X448) && (available_key_agreements&BCTBX_KEM_KYBER1024)) {
 			availableTypes[index] = ZRTP_KEYAGREEMENT_K448_KYB1024;
 			index++;
@@ -185,7 +211,7 @@ uint8_t bzrtpUtils_getAvailableCryptoTypes(uint8_t algoType, uint8_t availableTy
 		if (ret<7) {
 			return ret;
 		} else {
-			/* buggy behavior, we need this one for retrocompatibility, enfore its prensence at the end of the list */
+			/* buggy behavior, we need this one for retrocompatibility, enforce its presence at the end of the list */
 			availableTypes[6] = ZRTP_KEYAGREEMENT_Mult;
 			return 7;
 		}
@@ -258,7 +284,7 @@ int bzrtp_keyDerivationFunction(const uint8_t *key, const size_t keyLength,
 	index += contextLength;
 
 	/* end by L(hmacLength) in big-endian. hmacLength is in bytes and must be converted to bits before insertion in the text to hash */
-	input[index++] = 0; //(uint8_t)((hmacLength>>21)&0xFF); - hmacLength in bit cannot bit more than 2^11 so the first 2 bytes are always 0
+	input[index++] = 0; //(uint8_t)((hmacLength>>21)&0xFF); - hmacLength in bit cannot be more than 2^11 so the first 2 bytes are always 0
 	input[index++] = 0; //(uint8_t)((hmacLength>>13)&0xFF);
 	input[index++] = (uint8_t)((hmacLength>>5)&0xFF);
 	input[index++] = (uint8_t)((hmacLength<<3)&0xFF);
@@ -275,12 +301,12 @@ int bzrtp_keyDerivationFunction(const uint8_t *key, const size_t keyLength,
 void bzrtp_base32(uint32_t sas, char *output, int outputSize) {
 	int i, n, shift;
 
-	for (i=0,shift=27; i!=4; ++i,shift-=5) {
+	for (i=0,shift=27; i!=(outputSize-1); ++i,shift-=5) {
 		n = (sas>>shift) & 31;
 		output[i] = "ybndrfg8ejkmcpqxot1uwisza345h769"[n];
 	}
 
-	output[4] = '\0';
+	output[outputSize-1] = '\0';
 }
 
 /* Base256 function. Code from rfc section 5.1.6 */
@@ -288,8 +314,45 @@ void bzrtp_base256(uint32_t sas, char *output, int outputSize) {
 
 	// generate indexes and copy the appropriate words
 	int evenIndex = (sas >> 24) & 0xFF;
-	int oddIndex =  (sas >> 16) & 0xFF;
-	snprintf(output, outputSize, "%s:%s", pgpWordsEven[evenIndex], pgpWordsOdd[oddIndex]);
+	if (outputSize == 32) {
+		int oddIndex =  (sas >> 16) & 0xFF;
+		snprintf(output, outputSize, "%s:%s", pgpWordsEven[evenIndex], pgpWordsOdd[oddIndex]);
+	} else {
+		snprintf(output, outputSize, "%s", pgpWordsEven[evenIndex]);
+	}
+}
+
+void bzrtp_generate_incorrect_sas(uint32_t sas, char **incorrectSas, uint8_t sasAlgo) {
+	int nbUsedBits;
+	if(sasAlgo == ZRTP_SAS_B32) {
+		nbUsedBits = 10;
+	} else {
+		nbUsedBits = 8;
+	}
+	uint32_t sasFirstPart = (sas >> (32 - nbUsedBits)) & 0x3ff;
+	uint32_t sasSecondPart = (sas >> (32 - (nbUsedBits*2))) & 0x3ff;
+	uint32_t randIntTab[3];
+	for (int i = 0; i < 3; i++) {
+		uint32_t randInt;
+		bool_t sameAsOtherIncorrectSas;
+		do {
+			sameAsOtherIncorrectSas = FALSE;
+			randInt = bctbx_random();
+			for (int j = 0; j < i; j++) {
+				if(((randIntTab[j]>>(32-nbUsedBits))&0x3ff) == ((randInt>>(32-nbUsedBits))&0x3ff)){
+					sameAsOtherIncorrectSas = TRUE;
+				}
+			}
+		} while ((((randInt>>(32-nbUsedBits))&0x3ff) == sasFirstPart) || (((randInt>>(32-nbUsedBits))&0x3ff) == sasSecondPart) || sameAsOtherIncorrectSas);
+		randIntTab[i] = randInt;
+		if (sasAlgo == ZRTP_SAS_B32) {
+			incorrectSas[i] = (char *)bctbx_malloc((size_t)3*sizeof(char));
+			bzrtp_base32(randInt, (char*)incorrectSas[i], 3);
+		} else {
+			incorrectSas[i] = (char *)bctbx_malloc((size_t)16*sizeof(char));
+			bzrtp_base256(randInt, (char*)incorrectSas[i], 16);
+		}
+	}
 }
 
 uint32_t CRC32LookupTable[256] = {
@@ -394,6 +457,9 @@ uint32_t bzrtp_CRC32(uint8_t *input, uint16_t length) {
 
 bool_t bzrtp_isPostQuantum(uint8_t keyAgreementAlgo) {
 	switch (keyAgreementAlgo) {
+	case ZRTP_KEYAGREEMENT_MLK1:
+	case ZRTP_KEYAGREEMENT_MLK2:
+	case ZRTP_KEYAGREEMENT_MLK3:
 	case ZRTP_KEYAGREEMENT_KYB1:
 	case ZRTP_KEYAGREEMENT_KYB2:
 	case ZRTP_KEYAGREEMENT_KYB3:
@@ -401,8 +467,10 @@ bool_t bzrtp_isPostQuantum(uint8_t keyAgreementAlgo) {
 	case ZRTP_KEYAGREEMENT_HQC2:
 	case ZRTP_KEYAGREEMENT_HQC3:
 	case ZRTP_KEYAGREEMENT_K255_KYB512:
+	case ZRTP_KEYAGREEMENT_K255_MLK512:
 	case ZRTP_KEYAGREEMENT_K255_HQC128:
 	case ZRTP_KEYAGREEMENT_K448_KYB1024:
+	case ZRTP_KEYAGREEMENT_K448_MLK1024:
 	case ZRTP_KEYAGREEMENT_K448_HQC256:
 	case ZRTP_KEYAGREEMENT_K255_KYB512_HQC128:
 	case ZRTP_KEYAGREEMENT_K448_KYB1024_HQC256:
@@ -648,6 +716,9 @@ int bzrtp_updateCryptoFunctionPointers(bzrtpChannelContext_t *zrtpChannelContext
 	case ZRTP_KEYAGREEMENT_X448:
 	case ZRTP_KEYAGREEMENT_K255:
 	case ZRTP_KEYAGREEMENT_K448:
+	case ZRTP_KEYAGREEMENT_MLK1:
+	case ZRTP_KEYAGREEMENT_MLK2:
+	case ZRTP_KEYAGREEMENT_MLK3:
 	case ZRTP_KEYAGREEMENT_KYB1:
 	case ZRTP_KEYAGREEMENT_KYB2:
 	case ZRTP_KEYAGREEMENT_KYB3:
@@ -655,8 +726,10 @@ int bzrtp_updateCryptoFunctionPointers(bzrtpChannelContext_t *zrtpChannelContext
 	case ZRTP_KEYAGREEMENT_HQC2:
 	case ZRTP_KEYAGREEMENT_HQC3:
 	case ZRTP_KEYAGREEMENT_K255_KYB512:
+	case ZRTP_KEYAGREEMENT_K255_MLK512:
 	case ZRTP_KEYAGREEMENT_K255_HQC128:
 	case ZRTP_KEYAGREEMENT_K448_KYB1024:
+	case ZRTP_KEYAGREEMENT_K448_MLK1024:
 	case ZRTP_KEYAGREEMENT_K448_HQC256:
 	case ZRTP_KEYAGREEMENT_K255_KYB512_HQC128:
 	case ZRTP_KEYAGREEMENT_K448_KYB1024_HQC256:
@@ -880,6 +953,12 @@ uint8_t bzrtp_cryptoAlgoTypeStringToInt(uint8_t algoType[4], uint8_t algoFamily)
 			return ZRTP_KEYAGREEMENT_KYB2;
 		} else if (memcmp(algoType, "KYB3", 4) == 0) {
 			return ZRTP_KEYAGREEMENT_KYB3;
+		} else if (memcmp(algoType, "MLK1", 4) == 0) {
+			return ZRTP_KEYAGREEMENT_MLK1;
+		} else if (memcmp(algoType, "MLK2", 4) == 0) {
+			return ZRTP_KEYAGREEMENT_MLK2;
+		} else if (memcmp(algoType, "MLK3", 4) == 0) {
+			return ZRTP_KEYAGREEMENT_MLK3;
 		// HQC implementation used to be the round 3 submission (2020/10/11)
 		// -> it advertised itself as HQB<X> (and XKQ<x> for the X/Kyber/HQC hybrids
 		// Current HQC implementation is the one from round 4 submission (2024/02/23), incompatible with the previous one
@@ -892,8 +971,12 @@ uint8_t bzrtp_cryptoAlgoTypeStringToInt(uint8_t algoType[4], uint8_t algoFamily)
 			return ZRTP_KEYAGREEMENT_HQC3;
 		} else if (memcmp(algoType, "X1K1", 4) == 0) {
 			return ZRTP_KEYAGREEMENT_K255_KYB512;
+		} else if (memcmp(algoType, "X1M1", 4) == 0) {
+			return ZRTP_KEYAGREEMENT_K255_MLK512;
 		} else if (memcmp(algoType, "X1H1", 4) == 0) {
 			return ZRTP_KEYAGREEMENT_K255_HQC128;
+		} else if (memcmp(algoType, "X3M3", 4) == 0) {
+			return ZRTP_KEYAGREEMENT_K448_MLK1024;
 		} else if (memcmp(algoType, "X3K3", 4) == 0) {
 			return ZRTP_KEYAGREEMENT_K448_KYB1024;
 		} else if (memcmp(algoType, "X3H3", 4) == 0) {
@@ -1018,6 +1101,15 @@ void bzrtp_cryptoAlgoTypeIntToString(uint8_t algoTypeInt, uint8_t algoTypeString
 	case ZRTP_KEYAGREEMENT_KYB3:
 		memcpy(algoTypeString, "KYB3", 4);
 		break;
+	case ZRTP_KEYAGREEMENT_MLK1:
+		memcpy(algoTypeString, "MLK1", 4);
+		break;
+	case ZRTP_KEYAGREEMENT_MLK2:
+		memcpy(algoTypeString, "MLK2", 4);
+		break;
+	case ZRTP_KEYAGREEMENT_MLK3:
+		memcpy(algoTypeString, "MLK3", 4);
+		break;
 	// HQC implementation used to be the round 3 submission (2020/10/11)
 	// -> it advertised itself as HQB<X> (and XKQ<x> for the X/Kyber/HQC hybrids
 	// Current HQC implementation is the one from round 4 submission (2024/02/23), incompatible with the previous one
@@ -1031,12 +1123,18 @@ void bzrtp_cryptoAlgoTypeIntToString(uint8_t algoTypeInt, uint8_t algoTypeString
 	case ZRTP_KEYAGREEMENT_HQC3:
 		memcpy(algoTypeString, "HQD3", 4);
 		break;
+	case ZRTP_KEYAGREEMENT_K255_MLK512:
+		memcpy(algoTypeString, "X1M1", 4);
+		break;
 	case ZRTP_KEYAGREEMENT_K255_KYB512:
 		memcpy(algoTypeString, "X1K1", 4);
 		break;
 	case ZRTP_KEYAGREEMENT_K255_HQC128:
 		memcpy(algoTypeString, "X1H1", 4);
 		break;
+	case ZRTP_KEYAGREEMENT_K448_MLK1024:
+		memcpy(algoTypeString, "X3M3", 4);
+		break;
 	case ZRTP_KEYAGREEMENT_K448_KYB1024:
 		memcpy(algoTypeString, "X3K3", 4);
 		break;
@@ -1111,17 +1209,23 @@ void bzrtp_destroyKeyMaterial(bzrtpContext_t *zrtpContext, bzrtpChannelContext_t
 	bzrtp_DestroyKey(zrtpChannelContext->srtpSecrets.peerSrtpSalt, zrtpChannelContext->srtpSecrets.peerSrtpSaltLength, zrtpContext->RNGContext);
 	bzrtp_DestroyKey((uint8_t *)zrtpChannelContext->srtpSecrets.sas, zrtpChannelContext->srtpSecrets.sasLength, zrtpContext->RNGContext);
 
+
 	free(zrtpChannelContext->srtpSecrets.selfSrtpKey);
 	free(zrtpChannelContext->srtpSecrets.selfSrtpSalt);
 	free(zrtpChannelContext->srtpSecrets.peerSrtpKey);
 	free(zrtpChannelContext->srtpSecrets.peerSrtpSalt);
 	free(zrtpChannelContext->srtpSecrets.sas);
 
+	for (size_t i = 0; i < 3; i++) {
+		bctbx_free(zrtpChannelContext->srtpSecrets.incorrectSas[i]);
+		zrtpChannelContext->srtpSecrets.incorrectSas[i] = NULL;
+	}
+
 	zrtpChannelContext->srtpSecrets.selfSrtpKey=NULL;
 	zrtpChannelContext->srtpSecrets.selfSrtpSalt=NULL;
 	zrtpChannelContext->srtpSecrets.peerSrtpKey=NULL;
 	zrtpChannelContext->srtpSecrets.peerSrtpSalt=NULL;
-	zrtpChannelContext->srtpSecrets.sas=NULL;
+	zrtpChannelContext->srtpSecrets.sas = NULL;
 }
 
 /**
@@ -1130,37 +1234,47 @@ void bzrtp_destroyKeyMaterial(bzrtpContext_t *zrtpContext, bzrtpChannelContext_t
  *
  * @return The public key size in bytes
  */
-static uint16_t bzrt_getKEMPublicKeyLength(uint8_t keyAgreementAlgo) {
+static uint16_t bzrtp_getKEMPublicKeyLength(uint8_t keyAgreementAlgo) {
 #ifdef HAVE_BCTBXPQ
 	switch (keyAgreementAlgo) {
+	case ZRTP_KEYAGREEMENT_MLK1:
+		return bctoolbox::MLKEM512::kPkSize;
+	case ZRTP_KEYAGREEMENT_MLK2:
+		return bctoolbox::MLKEM768::kPkSize;
+	case ZRTP_KEYAGREEMENT_MLK3:
+		return bctoolbox::MLKEM1024::kPkSize;
 	case ZRTP_KEYAGREEMENT_KYB1:
-		return bctoolbox::KYBER512::pkSize;
+		return bctoolbox::KYBER512::kPkSize;
 	case ZRTP_KEYAGREEMENT_KYB2:
-		return bctoolbox::KYBER768::pkSize;
+		return bctoolbox::KYBER768::kPkSize;
 	case ZRTP_KEYAGREEMENT_KYB3:
-		return bctoolbox::KYBER1024::pkSize;
+		return bctoolbox::KYBER1024::kPkSize;
 	case ZRTP_KEYAGREEMENT_HQC1:
-		return bctoolbox::HQC128::pkSize;
+		return bctoolbox::HQC128::kPkSize;
 	case ZRTP_KEYAGREEMENT_HQC2:
-		return bctoolbox::HQC192::pkSize;
+		return bctoolbox::HQC192::kPkSize;
 	case ZRTP_KEYAGREEMENT_HQC3:
-		return bctoolbox::HQC256::pkSize;
+		return bctoolbox::HQC256::kPkSize;
 	case ZRTP_KEYAGREEMENT_K255:
-		return bctoolbox::K25519::pkSize;
+		return bctoolbox::K25519::kPkSize;
 	case ZRTP_KEYAGREEMENT_K448:
-		return bctoolbox::K448::pkSize;
+		return bctoolbox::K448::kPkSize;
+	case ZRTP_KEYAGREEMENT_K255_MLK512:
+		return bctoolbox::K25519::kPkSize + bctoolbox::MLKEM512::kPkSize;
 	case ZRTP_KEYAGREEMENT_K255_KYB512:
-		return bctoolbox::K25519::pkSize + bctoolbox::KYBER512::pkSize;
+		return bctoolbox::K25519::kPkSize + bctoolbox::KYBER512::kPkSize;
 	case ZRTP_KEYAGREEMENT_K255_HQC128:
-		return bctoolbox::K25519::pkSize + bctoolbox::HQC128::pkSize;
+		return bctoolbox::K25519::kPkSize + bctoolbox::HQC128::kPkSize;
+	case ZRTP_KEYAGREEMENT_K448_MLK1024:
+		return bctoolbox::K448::kPkSize + bctoolbox::MLKEM1024::kPkSize;
 	case ZRTP_KEYAGREEMENT_K448_KYB1024:
-		return bctoolbox::K448::pkSize + bctoolbox::KYBER1024::pkSize;
+		return bctoolbox::K448::kPkSize + bctoolbox::KYBER1024::kPkSize;
 	case ZRTP_KEYAGREEMENT_K448_HQC256:
-		return bctoolbox::K448::pkSize + bctoolbox::HQC256::pkSize;
+		return bctoolbox::K448::kPkSize + bctoolbox::HQC256::kPkSize;
 	case ZRTP_KEYAGREEMENT_K255_KYB512_HQC128:
-		return bctoolbox::K25519::pkSize + bctoolbox::KYBER512::pkSize + bctoolbox::HQC128::pkSize;
+		return bctoolbox::K25519::kPkSize + bctoolbox::KYBER512::kPkSize + bctoolbox::HQC128::kPkSize;
 	case ZRTP_KEYAGREEMENT_K448_KYB1024_HQC256:
-		return bctoolbox::K448::pkSize + bctoolbox::KYBER1024::pkSize + bctoolbox::HQC256::pkSize;
+		return bctoolbox::K448::kPkSize + bctoolbox::KYBER1024::kPkSize + bctoolbox::HQC256::kPkSize;
 	default:
 		return 0;
 	}
@@ -1175,37 +1289,47 @@ static uint16_t bzrt_getKEMPublicKeyLength(uint8_t keyAgreementAlgo) {
  *
  * @return The cipher text size in bytes
  */
-static uint16_t bzrt_getKEMCipherTextLength(uint8_t keyAgreementAlgo) {
+static uint16_t bzrtp_getKEMCipherTextLength(uint8_t keyAgreementAlgo) {
 #ifdef HAVE_BCTBXPQ
 	switch (keyAgreementAlgo) {
+	case ZRTP_KEYAGREEMENT_MLK1:
+		return bctoolbox::MLKEM512::kCtSize;
+	case ZRTP_KEYAGREEMENT_MLK2:
+		return bctoolbox::MLKEM768::kCtSize;
+	case ZRTP_KEYAGREEMENT_MLK3:
+		return bctoolbox::MLKEM1024::kCtSize;
 	case ZRTP_KEYAGREEMENT_KYB1:
-		return bctoolbox::KYBER512::ctSize;
+		return bctoolbox::KYBER512::kCtSize;
 	case ZRTP_KEYAGREEMENT_KYB2:
-		return bctoolbox::KYBER768::ctSize;
+		return bctoolbox::KYBER768::kCtSize;
 	case ZRTP_KEYAGREEMENT_KYB3:
-		return bctoolbox::KYBER1024::ctSize;
+		return bctoolbox::KYBER1024::kCtSize;
 	case ZRTP_KEYAGREEMENT_HQC1:
-		return bctoolbox::HQC128::ctSize;
+		return bctoolbox::HQC128::kCtSize;
 	case ZRTP_KEYAGREEMENT_HQC2:
-		return bctoolbox::HQC192::ctSize;
+		return bctoolbox::HQC192::kCtSize;
 	case ZRTP_KEYAGREEMENT_HQC3:
-		return bctoolbox::HQC256::ctSize;
+		return bctoolbox::HQC256::kCtSize;
 	case ZRTP_KEYAGREEMENT_K255:
-		return bctoolbox::K25519::ctSize;
+		return bctoolbox::K25519::kCtSize;
 	case ZRTP_KEYAGREEMENT_K448:
-		return bctoolbox::K448::ctSize;
+		return bctoolbox::K448::kCtSize;
+	case ZRTP_KEYAGREEMENT_K255_MLK512:
+		return bctoolbox::K25519::kCtSize + bctoolbox::MLKEM512::kCtSize;
 	case ZRTP_KEYAGREEMENT_K255_KYB512:
-		return bctoolbox::K25519::ctSize + bctoolbox::KYBER512::ctSize;
+		return bctoolbox::K25519::kCtSize + bctoolbox::KYBER512::kCtSize;
 	case ZRTP_KEYAGREEMENT_K255_HQC128:
-		return bctoolbox::K25519::ctSize + bctoolbox::HQC128::ctSize;
+		return bctoolbox::K25519::kCtSize + bctoolbox::HQC128::kCtSize;
+	case ZRTP_KEYAGREEMENT_K448_MLK1024:
+		return bctoolbox::K448::kCtSize + bctoolbox::MLKEM1024::kCtSize;
 	case ZRTP_KEYAGREEMENT_K448_KYB1024:
-		return bctoolbox::K448::ctSize + bctoolbox::KYBER1024::ctSize;
+		return bctoolbox::K448::kCtSize + bctoolbox::KYBER1024::kCtSize;
 	case ZRTP_KEYAGREEMENT_K448_HQC256:
-		return bctoolbox::K448::ctSize + bctoolbox::HQC256::ctSize;
+		return bctoolbox::K448::kCtSize + bctoolbox::HQC256::kCtSize;
 	case ZRTP_KEYAGREEMENT_K255_KYB512_HQC128:
-		return bctoolbox::K25519::ctSize + bctoolbox::KYBER512::ctSize + bctoolbox::HQC128::ctSize;
+		return bctoolbox::K25519::kCtSize + bctoolbox::KYBER512::kCtSize + bctoolbox::HQC128::kCtSize;
 	case ZRTP_KEYAGREEMENT_K448_KYB1024_HQC256:
-		return bctoolbox::K448::ctSize + bctoolbox::KYBER1024::ctSize + bctoolbox::HQC256::ctSize;
+		return bctoolbox::K448::kCtSize + bctoolbox::KYBER1024::kCtSize + bctoolbox::HQC256::kCtSize;
 	default:
 		return 0;
 	}
@@ -1235,7 +1359,7 @@ uint16_t bzrtp_computeKeyAgreementPublicValueLength(uint8_t keyAgreementAlgo, ui
 		switch (messageType) {
 		case MSGTYPE_COMMIT: // Commit packet includes the public key
 		{
-			auto ret = bzrt_getKEMPublicKeyLength(keyAgreementAlgo);
+			auto ret = bzrtp_getKEMPublicKeyLength(keyAgreementAlgo);
 			if (ret%4) { // The size MUST be a multiple of 4 as ZRTP nessage size is given in 4 bytes words not in bytes, it will just be padded at the end with 0 if needed
 				ret += 4-ret%4;
 			}
@@ -1243,7 +1367,7 @@ uint16_t bzrtp_computeKeyAgreementPublicValueLength(uint8_t keyAgreementAlgo, ui
 		}
 		case MSGTYPE_DHPART1: // DHPart1 packet holds the cipher text
 		{
-			auto ret = bzrt_getKEMCipherTextLength(keyAgreementAlgo);
+			auto ret = bzrtp_getKEMCipherTextLength(keyAgreementAlgo);
 			if (ret%4) { // The size MUST be a multiple of 4 as ZRTP nessage size is given in 4 bytes words not in bytes, it will just be padded at the end with 0 if needed
 				ret += 4-ret%4;
 			}
@@ -1274,25 +1398,33 @@ uint16_t bzrtp_computeKeyAgreementSharedSecretLength(uint8_t keyAgreementAlgo, u
 	case ZRTP_KEYAGREEMENT_EC52 :
 		return 132;
 #ifdef HAVE_BCTBXPQ
+	case ZRTP_KEYAGREEMENT_MLK1:
+		return bctoolbox::MLKEM512::kSsSize;
+	case ZRTP_KEYAGREEMENT_MLK2:
+		return bctoolbox::MLKEM768::kSsSize;
+	case ZRTP_KEYAGREEMENT_MLK3:
+		return bctoolbox::MLKEM1024::kSsSize;
 	case ZRTP_KEYAGREEMENT_KYB1:
-		return bctoolbox::KYBER512::ssSize;
+		return bctoolbox::KYBER512::kSsSize;
 	case ZRTP_KEYAGREEMENT_KYB2:
-		return bctoolbox::KYBER768::ssSize;
+		return bctoolbox::KYBER768::kSsSize;
 	case ZRTP_KEYAGREEMENT_KYB3:
-		return bctoolbox::KYBER1024::ssSize;
+		return bctoolbox::KYBER1024::kSsSize;
 	case ZRTP_KEYAGREEMENT_HQC1:
-		return bctoolbox::HQC128::ssSize;
+		return bctoolbox::HQC128::kSsSize;
 	case ZRTP_KEYAGREEMENT_HQC2:
-		return bctoolbox::HQC192::ssSize;
+		return bctoolbox::HQC192::kSsSize;
 	case ZRTP_KEYAGREEMENT_HQC3:
-		return bctoolbox::HQC256::ssSize;
+		return bctoolbox::HQC256::kSsSize;
 	case ZRTP_KEYAGREEMENT_K255:
-		return bctoolbox::K25519::ssSize;
+		return bctoolbox::K25519::kSsSize;
 	case ZRTP_KEYAGREEMENT_K448:
-		return bctoolbox::K448::ssSize;
+		return bctoolbox::K448::kSsSize;
 	case ZRTP_KEYAGREEMENT_K255_KYB512:
+	case ZRTP_KEYAGREEMENT_K255_MLK512:
 	case ZRTP_KEYAGREEMENT_K255_HQC128:
 	case ZRTP_KEYAGREEMENT_K448_KYB1024:
+	case ZRTP_KEYAGREEMENT_K448_MLK1024:
 	case ZRTP_KEYAGREEMENT_K448_HQC256:
 	case ZRTP_KEYAGREEMENT_K255_KYB512_HQC128:
 	case ZRTP_KEYAGREEMENT_K448_KYB1024_HQC256:
@@ -1305,6 +1437,9 @@ uint16_t bzrtp_computeKeyAgreementSharedSecretLength(uint8_t keyAgreementAlgo, u
 
 bool_t bzrtp_isKem(uint8_t keyAgreementAlgo) {
 	switch (keyAgreementAlgo) {
+	case ZRTP_KEYAGREEMENT_MLK1:
+	case ZRTP_KEYAGREEMENT_MLK2:
+	case ZRTP_KEYAGREEMENT_MLK3:
 	case ZRTP_KEYAGREEMENT_KYB1:
 	case ZRTP_KEYAGREEMENT_KYB2:
 	case ZRTP_KEYAGREEMENT_KYB3:
@@ -1314,8 +1449,10 @@ bool_t bzrtp_isKem(uint8_t keyAgreementAlgo) {
 	case ZRTP_KEYAGREEMENT_K255:
 	case ZRTP_KEYAGREEMENT_K448:
 	case ZRTP_KEYAGREEMENT_K255_KYB512:
+	case ZRTP_KEYAGREEMENT_K255_MLK512:
 	case ZRTP_KEYAGREEMENT_K255_HQC128:
 	case ZRTP_KEYAGREEMENT_K448_KYB1024:
+	case ZRTP_KEYAGREEMENT_K448_MLK1024:
 	case ZRTP_KEYAGREEMENT_K448_HQC256:
 	case ZRTP_KEYAGREEMENT_K255_KYB512_HQC128:
 	case ZRTP_KEYAGREEMENT_K448_KYB1024_HQC256:
@@ -1329,7 +1466,7 @@ uint16_t bzrtp_computeCommitMessageVariableLength(uint8_t keyAgreementAlgo) {
 	if (keyAgreementAlgo == ZRTP_KEYAGREEMENT_Prsh) return 24; /* nonce (16 bytes) and keyID(8 bytes) are 24 bytes length in preshared Commit message format */
 	if (keyAgreementAlgo == ZRTP_KEYAGREEMENT_Mult) return 16; /* nonce is 16 bytes length in multistream Commit message format */
 	if (bzrtp_isKem(keyAgreementAlgo)) {
-		auto ret = bzrt_getKEMPublicKeyLength(keyAgreementAlgo);
+		auto ret = bzrtp_getKEMPublicKeyLength(keyAgreementAlgo);
 		if (ret%4) { // The size MUST be a multiple of 4 as ZRTP nessage size is given in 4 bytes words not in bytes, it will just be padded at the end with 0 if needed
 			ret += 4-ret%4;
 		}
@@ -1391,6 +1528,15 @@ bzrtp_KEMContext_t *bzrtp_createKEMContext(uint8_t keyAgreementAlgo, uint8_t has
 	int hashId = bzrtp_getHashAlgoId(hashAlgo);
 
 	switch (keyAgreementAlgo) {
+	case ZRTP_KEYAGREEMENT_MLK1:
+		context->ctx = std::make_shared<bctoolbox::MLKEM512>();
+		break;
+	case ZRTP_KEYAGREEMENT_MLK2:
+		context->ctx = std::make_shared<bctoolbox::MLKEM768>();
+		break;
+	case ZRTP_KEYAGREEMENT_MLK3:
+		context->ctx = std::make_shared<bctoolbox::MLKEM1024>();
+		break;
 	case ZRTP_KEYAGREEMENT_KYB1:
 		context->ctx = std::make_shared<bctoolbox::KYBER512>();
 		break;
@@ -1415,12 +1561,18 @@ bzrtp_KEMContext_t *bzrtp_createKEMContext(uint8_t keyAgreementAlgo, uint8_t has
 	case ZRTP_KEYAGREEMENT_K448:
 		context->ctx = std::make_shared<bctoolbox::K448>(hashId);
 		break;
+	case ZRTP_KEYAGREEMENT_K255_MLK512:
+		context->ctx = std::make_shared<bctoolbox::HYBRID_KEM>(std::list<std::shared_ptr<bctoolbox::KEM>>({std::make_shared<bctoolbox::K25519>(hashId), std::make_shared<bctoolbox::MLKEM512>()}), hashId);
+		break;
 	case ZRTP_KEYAGREEMENT_K255_KYB512:
 		context->ctx = std::make_shared<bctoolbox::HYBRID_KEM>(std::list<std::shared_ptr<bctoolbox::KEM>>({std::make_shared<bctoolbox::K25519>(hashId), std::make_shared<bctoolbox::KYBER512>()}), hashId);
 		break;
 	case ZRTP_KEYAGREEMENT_K255_HQC128:
 		context->ctx = std::make_shared<bctoolbox::HYBRID_KEM>(std::list<std::shared_ptr<bctoolbox::KEM>>({std::make_shared<bctoolbox::K25519>(hashId), std::make_shared<bctoolbox::HQC128>()}), hashId);
 		break;
+	case ZRTP_KEYAGREEMENT_K448_MLK1024:
+		context->ctx = std::make_shared<bctoolbox::HYBRID_KEM>(std::list<std::shared_ptr<bctoolbox::KEM>>({std::make_shared<bctoolbox::K448>(hashId), std::make_shared<bctoolbox::MLKEM1024>()}), hashId);
+		break;
 	case ZRTP_KEYAGREEMENT_K448_KYB1024:
 		context->ctx = std::make_shared<bctoolbox::HYBRID_KEM>(std::list<std::shared_ptr<bctoolbox::KEM>>({std::make_shared<bctoolbox::K448>(hashId), std::make_shared<bctoolbox::KYBER1024>()}), hashId);
 		break;
@@ -1449,7 +1601,7 @@ int bzrtp_KEM_generateKeyPair(bzrtp_KEMContext_t *ctx) {
 		return BZRTP_ERROR_CONTEXTNOTREADY;
 	}
 
-	return ctx->ctx->crypto_kem_keypair(ctx->publicKey, ctx->secretKey);
+	return ctx->ctx->keyGen(ctx->publicKey, ctx->secretKey);
 }
 
 /**
@@ -1465,7 +1617,7 @@ int bzrtp_KEM_getPublicKey(bzrtp_KEMContext_t *ctx, uint8_t *publicKey) {
 	}
 
 	if (ctx->publicKey.size() > 0) {
-		memcpy(publicKey,ctx->publicKey.data(), ctx->publicKey.size());
+		std::memcpy(publicKey,ctx->publicKey.data(), ctx->publicKey.size());
 		return 0;
 	}
 	return BZRTP_ERROR_CONTEXTNOTREADY;
@@ -1477,7 +1629,7 @@ int bzrtp_KEM_getSharedSecret(bzrtp_KEMContext_t *ctx, uint8_t *sharedSecret) {
 	}
 
 	if (ctx->sharedSecret.size() > 0) {
-		memcpy(sharedSecret,ctx->sharedSecret.data(), ctx->sharedSecret.size());
+		std::memcpy(sharedSecret,ctx->sharedSecret.data(), ctx->sharedSecret.size());
 		return 0;
 	}
 	return BZRTP_ERROR_CONTEXTNOTREADY;
@@ -1489,11 +1641,10 @@ int bzrtp_KEM_encaps(bzrtp_KEMContext_t *ctx, uint8_t *publicKey, uint8_t *ciphe
 	}
 
 	std::vector<uint8_t> ct;
-	std::vector<uint8_t> pk(ctx->ctx->get_pkSize());
-	memcpy(pk.data(), publicKey, pk.size());
+	std::vector<uint8_t> pk(publicKey, publicKey + ctx->ctx->getPkSize());
 
-	if (ctx->ctx->crypto_kem_enc(ct, ctx->sharedSecret, pk) == 0) {
-		memcpy(cipherText, ct.data(), ct.size());
+	if (ctx->ctx->encaps(ct, ctx->sharedSecret, pk) == 0) {
+		std::memcpy(cipherText, ct.data(), ct.size());
 		return 0;
 	}
 	return BZRTP_ERROR_CONTEXTNOTREADY;
@@ -1503,9 +1654,8 @@ int bzrtp_KEM_decaps(bzrtp_KEMContext_t *ctx, uint8_t *cipherText) {
 	if (ctx == NULL || ctx->ctx == NULL || ctx->secretKey.size()==0) {
 		return BZRTP_ERROR_CONTEXTNOTREADY;
 	}
-	std::vector<uint8_t> ct(ctx->ctx->get_ctSize());
-	memcpy(ct.data(), cipherText, ct.size());
-	if (ctx->ctx->crypto_kem_dec(ctx->sharedSecret, ct, ctx->secretKey) == 0) {
+	std::vector<uint8_t> ct(cipherText, cipherText + ctx->ctx->getCtSize());
+	if (ctx->ctx->decaps(ctx->sharedSecret, ct, ctx->secretKey) == 0) {
 		return 0;
 	}
 	return BZRTP_ERROR_CONTEXTNOTREADY;
